# Copyright (c) 2020, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Section functions and functional tools'''

import numpy as np

from neurom import morphmath as mm
from neurom.core.dataformat import COLS
from neurom._compat import range
from neurom.morphmath import interval_lengths


def section_path_length(section):
    '''Path length from section to root'''
    return sum(s.length for s in section.iupstream())


def section_volume(section):
    '''Volume of a section'''
    return section.volume


def section_area(section):
    '''Surface area of a section'''
    return section.area


def section_tortuosity(section):
    '''Tortuosity of a section

    The tortuosity is defined as the ratio of the path length of a section
    and the euclidian distnce between its end points.

    The path length is the sum of distances between consecutive points.

    If the section contains less than 2 points, the value 1 is returned.
    '''
    pts = section.points
    return 1 if len(pts) < 2 else mm.section_length(pts) / mm.point_dist(pts[-1], pts[0])


def section_end_distance(section):
    '''End to end distance of a section

    The end to end distance of a section is defined as
    the euclidian distnce between its end points.

    If the section contains less than 2 points, the value 0 is returned.
    '''
    pts = section.points
    return 0 if len(pts) < 2 else mm.point_dist(pts[-1], pts[0])


def branch_order(section):
    '''Branching order of a tree section

    The branching order is defined as the depth of the tree section.

    Note:
        The first level has branch order 1.
    '''
    return sum(1 for _ in section.iupstream()) - 1


def segment_lengths(section):
    '''Returns the list of segment lengths within the section'''
    return interval_lengths(section.points)


def section_radial_distance(section, origin):
    '''Return the radial distances of a tree section to a given origin point

    The radial distance is the euclidian distance between the
    end-point point of the section and the origin point in question.

    Parameters:
        section: neurite section object
        origin: point to which distances are measured. It must have at least 3\
            components. The first 3 components are (x, y, z).
    '''
    return mm.point_dist(section.points[-1], origin)


def section_meander_angles(section):
    '''Inter-segment opening angles in a section'''
    p = section.points
    return [mm.angle_3points(p[i - 1], p[i - 2], p[i])
            for i in range(2, len(p))]


def strahler_order(section):
    '''Branching order of a tree section

    The strahler order is the inverse of the branch order,
    since this is computed from the tips of the tree
    towards the root.

    This implementation is a translation of the three steps described in
    Wikipedia (https://en.wikipedia.org/wiki/Strahler_number):

       - If the node is a leaf (has no children), its Strahler number is one.
       - If the node has one child with Strahler number i, and all other children
         have Strahler numbers less than i, then the Strahler number of the node
         is i again.
       - If the node has two or more children with Strahler number i, and no
         children with greater number, then the Strahler number of the node is
         i + 1.

    No efforts have been invested in making it computationnaly efficient, but
    it computes acceptably fast on tested morphologies (i.e., no waiting time).
    '''
    if section.children:
        child_orders = [strahler_order(child) for child in section.children]
        max_so_children = max(child_orders)
        it = iter(co == max_so_children for co in child_orders)
        #  check if there are *two* or more children w/ the max_so_children
        any(it)
        if any(it):
            return max_so_children + 1
        return max_so_children
    return 1


def locate_segment_position(section, fraction):
    '''
    Segment ID / offset corresponding to a given fraction of section length.
    '''
    return mm.path_fraction_id_offset(section.points, fraction)


def section_mean_radius(section):
    '''Compute the mean radius of a section weighted by segment lengths'''
    radii = section.points[:, COLS.R]
    points = section.points[:, COLS.XYZ]
    lengths = np.linalg.norm(points[1:] - points[:-1], axis=1)
    mean_radii = 0.5 * (radii[1:] + radii[:-1])
    return np.sum(mean_radii * lengths) / np.sum(lengths)


def downstream_pathlength(section):
    '''Compute the total downstream length starting from a section'''
    return sum(sec.length for sec in section.ipreorder())



# Copyright (c) 2016, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Reader for Neurolucida .ASC files, v3, reversed engineered from looking at output from
Neuroludica
'''

import warnings
from io import open

import numpy as np

from neurom._compat import StringType
from neurom.core.dataformat import COLS, POINT_TYPE

from .datawrapper import DataWrapper

WANTED_SECTIONS = {
    'CellBody': POINT_TYPE.SOMA,
    'Axon': POINT_TYPE.AXON,
    'Dendrite': POINT_TYPE.BASAL_DENDRITE,
    'Apical': POINT_TYPE.APICAL_DENDRITE,
}
UNWANTED_SECTION_NAMES = [
    # Meta-data?
    'Closed', 'Color', 'FillDensity', 'GUID', 'ImageCoords', 'MBFObjectType',
    'Marker', 'Name', 'Resolution', 'Set', 'Description',
    # Marker names?
    'Asterisk', 'Cross', 'Dot', 'DoubleCircle', 'FilledCircle', 'FilledDownTriangle',
    'FilledSquare', 'FilledStar', 'FilledUpTriangle', 'FilledUpTriangle', 'Flower',
    'Flower2', 'OpenCircle', 'OpenDiamond', 'OpenDownTriangle', 'OpenSquare', 'OpenStar',
    'OpenUpTriangle', 'Plus', 'ShadedStar', 'Splat', 'TriStar',
]
UNWANTED_SECTIONS = {name: True for name in UNWANTED_SECTION_NAMES}


def _match_section(section, match):
    '''checks whether the `type` of section is in the `match` dictionary

    Works around the unknown ordering of s-expressions in each section.
    For instance, the `type` is the 3-rd one in for CellBodies
        ("CellBody"
         (Color Yellow)
         (CellBody)
         (Set "cell10")
        )

    Returns:
        value associated with match[section_type], None if no match
    '''
    # TODO: rewrite this so it is more clear, and handles sets & dictionaries for matching
    for i in range(5):
        if i >= len(section):
            return None
        if isinstance(section[i], StringType) and section[i] in match:
            return match[section[i]]
    return None


def _get_tokens(morph_fd):
    '''split a file-like into tokens: split on whitespace

    Note: this also strips newlines and comments
    '''
    for line in morph_fd:
        line = line.rstrip()   # remove \r\n
        line = line.split(';', 1)[0]  # strip comments
        squash_token = []  # quoted strings get squashed into one token

        if '<(' in line:  # skip spines, which exist on a single line
            assert ')>' in line, 'Missing end of spine'
            # The following line is covered but 'tox -e coverage does not see it'
            # TODO: find out why
            continue  # pragma: no cover

        for token in line.replace('(', ' ( ').replace(')', ' ) ').split():
            if squash_token:
                squash_token.append(token)
                if token.endswith('"'):
                    token = ' '.join(squash_token)
                    squash_token = []
                    yield token
            elif token.startswith('"') and not token.endswith('"'):
                squash_token.append(token)
            else:
                yield token


def _parse_section(token_iter):
    '''take a stream of tokens, and create the tree structure that is defined
    by the s-expressions
    '''
    sexp = []
    for token in token_iter:
        if token == '(':
            new_sexp = _parse_section(token_iter)
            if not _match_section(new_sexp, UNWANTED_SECTIONS):
                sexp.append(new_sexp)
        elif token == ')':
            return sexp
        else:
            sexp.append(token)
    return sexp


def _parse_sections(morph_fd):
    '''returns array of all the sections that exist

    The format is nested lists that correspond to the s-expressions
    '''
    sections = []
    token_iter = _get_tokens(morph_fd)
    for token in token_iter:
        if token == '(':  # find top-level sections
            section = _parse_section(token_iter)
            if not _match_section(section, UNWANTED_SECTIONS):
                sections.append(section)
    return sections


def _flatten_subsection(subsection, _type, offset, parent):
    '''Flatten a subsection from its nested version

    Args:
        subsection: Nested subsection as produced by _parse_section, except one level in
        _type: type of section, ie: AXON, etc
        parent: first element has this as it's parent
        offset: position in the final array of the first element

    Returns:
        Generator of values corresponding to [X, Y, Z, R, TYPE, ID, PARENT_ID]
    '''
    for row in subsection:
        # TODO: Figure out what these correspond to in neurolucida
        if row in ('Low', 'Generated', 'High', ):
            continue

        if isinstance(row[0], StringType):
            if len(row) in (4, 5, ):
                if len(row) == 5:
                    assert row[4][0] == 'S', \
                        'Only known usage of a fifth member is Sn, found: %s' % row[4][0]
                yield (float(row[0]), float(row[1]), float(row[2]), float(row[3]) / 2.,
                       _type, offset, parent)
                parent = offset
                offset += 1
        elif isinstance(row[0], list):
            split_parent = offset - 1
            start_offset = 0

            slices = []
            start = 0
            for i, value in enumerate(row):
                if value == '|':
                    slices.append(slice(start + start_offset, i))
                    start = i + 1
            slices.append(slice(start + start_offset, len(row)))

            for split_slice in slices:
                for _row in _flatten_subsection(row[split_slice], _type, offset,
                                                split_parent):
                    offset += 1
                    yield _row


def _extract_section(section):
    '''Find top level sections, and get their flat contents, and append them all

    Returns a numpy array with the row format:
        [X, Y, Z, R, TYPE, ID, PARENT_ID]

    Note: PARENT_ID starts at -1 for soma and 0 for neurites
    '''
    # sections with only one element will be skipped,
    if len(section) == 1:
        assert section[0] == 'Sections', \
            ('Only known usage of a single Section content is "Sections", found %s' %
             section[0])
        return None

    # try and detect type
    _type = WANTED_SECTIONS.get(section[0][0], None)

    start = 1
    # CellBody often has [['"CellBody"'], ['CellBody'] as its first two elements
    if _type is None:
        _type = WANTED_SECTIONS.get(section[1][0], None)

        if _type is None:  # can't determine the type
            return None
        start = 2

    parent = -1 if _type == POINT_TYPE.SOMA else 0
    subsections = list(_flatten_subsection(section[start:], _type, offset=0,
                                           parent=parent))

    return np.array(subsections)


def _sections_to_raw_data(sections):
    '''convert list of sections into the `raw_data` format used in neurom

    This finds the soma, and attaches the neurites
    '''
    soma = None
    neurites = []
    for section in sections:
        neurite = _extract_section(section)
        if neurite is None:
            continue

        if neurite[0][COLS.TYPE] == POINT_TYPE.SOMA:
            assert soma is None, 'Multiple somas defined in file'
            soma = neurite
        else:
            neurites.append(neurite)
    assert soma is not None, 'Missing CellBody element (ie. soma)'

    total_length = len(soma) + sum(len(neurite) for neurite in neurites)
    ret = np.zeros((total_length, 7,), dtype=np.float64)
    pos = len(soma)
    ret[0:pos, :] = soma

    for neurite in neurites:
        end = pos + len(neurite)
        ret[pos:end, :] = neurite
        ret[pos:end, COLS.P] += pos
        ret[pos:end, COLS.ID] += pos
        # TODO: attach the neurite at the closest point on the soma
        ret[pos, COLS.P] = len(soma) - 1
        pos = end

    return ret


def read(morph_file, data_wrapper=DataWrapper):
    '''return a 'raw_data' np.array with the full neuron, and the format of the file
    suitable to be wrapped by DataWrapper
    '''

    warnings.warn('This is an experimental reader. '
                  'There are no guarantees regarding ability to parse '
                  'Neurolucida .asc files or correctness of output.')

    with open(morph_file, encoding='utf-8', errors='replace') as morph_fd:
        sections = _parse_sections(morph_fd)
    raw_data = _sections_to_raw_data(sections)
    return data_wrapper(raw_data, 'NL-ASCII')



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Fast neuron IO module'''

from copy import deepcopy

import numpy as np

from neurom.core import (Section, Neurite, Neuron, NeuriteType, SomaError,)
from neurom.core.dataformat import POINT_TYPE, COLS, ROOT_ID
from neurom.core._soma import make_soma, SOMA_CONTOUR, SOMA_CYLINDER


class FstNeuron(Neuron):
    '''Class representing a neuron'''

    def __init__(self, data_wrapper, name='Neuron'):
        self._data = data_wrapper
        neurites, sections = make_neurites(self._data)
        soma_check, soma_class = _SOMA_CONFIG[self._data.fmt]
        soma = make_soma(self._data.soma_points(), soma_check, soma_class)
        super(FstNeuron, self).__init__(soma, neurites, sections, name)
        self._points = None

    @property
    def points(self):
        '''Return unordered array with all the points in this neuron'''
        if self._points is None:
            _points = self.soma.points.tolist()
            for n in self.neurites:
                _points.extend(n.points.tolist())
            self._points = np.array(_points)

        return self._points

    def transform(self, trans):
        '''Return a copy of this neuron with a 3D transformation applied'''
        _data = deepcopy(self._data)
        _data.data_block[:, 0:3] = trans(_data.data_block[:, 0:3])
        return FstNeuron(_data, self.name)

    def __deepcopy__(self, memo):
        '''Deep-copy neuron object

        Note:
            Efficient copying is performed by deep-copying the internal
            data block and building a neuron from it
        '''
        return FstNeuron(deepcopy(self._data, memo), self.name)


def make_neurites(rdw):
    '''Build neurite trees from a raw data wrapper'''
    post_action = _NEURITE_ACTION[rdw.fmt]
    trunks = rdw.neurite_root_section_ids()
    if not trunks:
        return [], []

    # One pass over sections to build nodes
    nodes = tuple(Section(section_id=i,
                          points=rdw.data_block[sec.ids],
                          section_type=_TREE_TYPES[sec.ntype])
                  for i, sec in enumerate(rdw.sections))

    # One pass over nodes to connect children to parents
    for i, node in enumerate(nodes):
        parent_id = rdw.sections[i].pid
        parent_type = nodes[parent_id].type
        # only connect neurites
        if parent_id != ROOT_ID and parent_type != NeuriteType.soma:
            nodes[parent_id].add_child(node)

    neurites = tuple(Neurite(nodes[i]) for i in trunks)

    if post_action is not None:
        for n in neurites:
            post_action(n.root_node)

    return neurites, nodes


def _remove_soma_initial_point(tree):
    '''Remove tree's initial point if soma'''
    if tree.points[0][COLS.TYPE] == POINT_TYPE.SOMA:
        tree.points = tree.points[1:]


def _check_soma_topology_swc(points):
    '''check if points form valid soma

    Currently checks if there are bifurcations within a soma
    with more than three points.
    '''
    if len(points) == 3:
        return

    parents = tuple(p[COLS.P] for p in points if p[COLS.P] != ROOT_ID)
    if len(parents) > len(set(parents)):
        raise SomaError("Bifurcating soma")


_TREE_TYPES = tuple(NeuriteType)

_NEURITE_ACTION = {
    'SWC': _remove_soma_initial_point,
    'H5V1': None,
    'H5V2': None,
    'NL-ASCII': _remove_soma_initial_point,
}

_SOMA_CONFIG = {
    # after much debate (https://github.com/BlueBrain/NeuroM/issues/597)
    # and research:
    #
    #  Cannon et al., 1998: http://www.sciencedirect.com/science/article/pii/S0165027098000910
    #    'Each line has the same seven fields: numbered point index, user defined flag denoting the
    #    specific part of the structure (cell body, dendrite, axon etc.), three-dimensional position
    #    (x, y and z, in mm), radius (r, in mm), and the parent point index'
    #
    #  Ascoli et al., 2001: http://www.jstor.org/stable/3067144
    #    'In the SWC format, dendritic segments are characterized by an identification number, a
    #    type (to distinguish basal, apical, proximal, distal and lateral trees), the x, y, z
    #    positions of the cylinder ending point (in pm with respect to a fixed reference point), a
    #    radius value (also in pm), and the identification number of the 'parent', i.e. the adjacent
    #    cylinder in the path to the soma (the parent of the root being the soma itself)."
    #
    # that the SWC format uses cylinders to represent the soma.
    'SWC': (_check_soma_topology_swc, SOMA_CYLINDER),
    'H5V1': (None, SOMA_CONTOUR),
    'H5V2': (None, SOMA_CONTOUR),
    'NL-ASCII': (None, SOMA_CONTOUR),
}



'''Test neurom.io.utils'''
import os
import numpy as np
from nose import tools as nt

from neurom.io import datawrapper as dw
from neurom.core.dataformat import POINT_TYPE, ROOT_ID


def test__merge_sections():
    default_sec = dw.DataBlockSection()

    sec_a = dw.DataBlockSection([], ntype=0, pid=-1)
    sec_b = dw.DataBlockSection([], ntype=0, pid=-1)
    dw._merge_sections(sec_a, sec_b)
    nt.eq_(sec_a, default_sec)

    sec_a = dw.DataBlockSection(range(10), ntype=1, pid=1)
    sec_b = dw.DataBlockSection(range(9, 20), ntype=10, pid=10)
    dw._merge_sections(sec_a, sec_b)
    nt.eq_(sec_a, default_sec)
    nt.eq_(sec_b.ids, list(range(20))) # Note: 9 is in this list from sec_a, not from sec_b
    nt.eq_(sec_b.ntype, 1)
    nt.eq_(sec_b.pid, 1)


#def test__section_end_points():
#    _section_end_points
#
#def test__extract_sections():
#    pass

#DataWrapper
#neurite_root_section_ids
#soma_points

def test_DataBlockSection_str():
    s = str(dw.DataBlockSection())
    nt.ok_('DataBlockSection' in s)

def test_BlockNeuronBuilder():
    builder = dw.BlockNeuronBuilder()
    soma_points = np.array([[0, 0, 0, 1]])
    builder.add_section(0, ROOT_ID, POINT_TYPE.SOMA, soma_points)
    wrapped = builder.get_datawrapper()
    nt.eq_(len(wrapped.data_block), 1)

    builder.add_section(1, 0, POINT_TYPE.AXON, np.array([[1, 0, 0, 1]]))
    wrapped = builder.get_datawrapper()
    nt.eq_(len(wrapped.data_block), 2)

    #add child before parent, also, don't have contiguous section numbers
    # meaning that the ids will be renumbered
    builder.add_section(10, 2, POINT_TYPE.APICAL_DENDRITE, np.array([[10, 0, 0, 1]]))
    builder.add_section(2, 0, POINT_TYPE.APICAL_DENDRITE, np.array([[2, 0, 0, 1]]))
    wrapped = builder.get_datawrapper()
    nt.eq_(len(wrapped.data_block), 4)
    np.testing.assert_allclose(
        wrapped.data_block,
        np.array([[ 0., 0., 0., 1., 1., 0., -1.],
                  [ 1., 0., 0., 1., 2., 1.,  0.],
                  [ 2., 0., 0., 1., 4., 2.,  0.],
                  [10., 0., 0., 1., 4., 3.,  2.]]))



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Module with consistency/validity checks for raw data  blocks'''
import numpy as np
from neurom.check import CheckResult
from neurom.core.dataformat import COLS
from neurom.core.dataformat import POINT_TYPE
from neurom.core import make_soma
from neurom.fst._core import make_neurites
from neurom.exceptions import SomaError


def has_sequential_ids(data_wrapper):
    '''Check that IDs are increasing and consecutive

    returns tuple (bool, list of IDs that are not consecutive
    with their predecessor)
    '''
    db = data_wrapper.data_block
    ids = db[:, COLS.ID]
    steps = ids[np.where(np.diff(ids) != 1)[0] + 1].astype(int)
    return CheckResult(len(steps) == 0, steps)


def no_missing_parents(data_wrapper):
    '''Check that all points have existing parents
    Point's parent ID must exist and parent must be declared
    before child.

    Returns:
        CheckResult with result and list of IDs that have no parent
    '''
    db = data_wrapper.data_block
    ids = np.setdiff1d(db[:, COLS.P], db[:, COLS.ID])[1:]
    return CheckResult(len(ids) == 0, ids.astype(np.int) + 1)


def is_single_tree(data_wrapper):
    '''Check that data forms a single tree

    Only the first point has ID of -1.

    Returns:
        CheckResult with result and list of IDs

    Note:
        This assumes no_missing_parents passed.
    '''
    db = data_wrapper.data_block
    bad_ids = db[db[:, COLS.P] == -1][1:, COLS.ID]
    return CheckResult(len(bad_ids) == 0, bad_ids.tolist())


def has_increasing_ids(data_wrapper):
    '''Check that IDs are increasing

    Returns:
        CheckResult with result and list of IDs that are inconsistent
        with their predecessor
    '''
    db = data_wrapper.data_block
    ids = db[:, COLS.ID]
    steps = ids[np.where(np.diff(ids) <= 0)[0] + 1].astype(int)
    return CheckResult(len(steps) == 0, steps)


def has_soma_points(data_wrapper):
    '''Checks if the TYPE column of raw data block has an element of type soma

    Returns:
        CheckResult with result
    '''
    db = data_wrapper.data_block
    return CheckResult(POINT_TYPE.SOMA in db[:, COLS.TYPE], None)


def has_all_finite_radius_neurites(data_wrapper, threshold=0.0):
    '''Check that all points with neurite type have a finite radius

    Returns:
        CheckResult with result and list of IDs of neurite points with zero radius
    '''
    db = data_wrapper.data_block
    neurite_ids = np.in1d(db[:, COLS.TYPE], POINT_TYPE.NEURITES)
    zero_radius_ids = db[:, COLS.R] <= threshold
    bad_pts = np.array(db[neurite_ids & zero_radius_ids][:, COLS.ID],
                       dtype=int).tolist()
    return CheckResult(len(bad_pts) == 0, bad_pts)


def has_valid_soma(data_wrapper):
    '''Check if a data block has a valid soma

    Returns:
        CheckResult with result
    '''
    try:
        make_soma(data_wrapper.soma_points())
        return CheckResult(True)
    except SomaError:
        return CheckResult(False)


def has_valid_neurites(data_wrapper):
    '''Check if any neurites can be reconstructed from data block

    Returns:
        CheckResult with result
    '''
    n, _ = make_neurites(data_wrapper)
    return CheckResult(len(n) > 0)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test neurom.features._core module'''

import os
from copy import deepcopy

import numpy as np
from nose import tools as nt

from neurom import io as _io
from neurom.fst import _core

_path = os.path.dirname(os.path.abspath(__file__))
DATA_ROOT = os.path.join(_path, '../../../test_data')
DATA_PATH = os.path.join(_path, '../../../test_data/valid_set')
FILENAMES = [os.path.join(DATA_PATH, f)
             for f in ['Neuron.swc', 'Neuron_h5v1.h5', 'Neuron_h5v2.h5']]


def test_neuron_name():

    d = _io.load_data(FILENAMES[0])
    nrn = _core.FstNeuron(d, '12af3rg')
    nt.eq_(nrn.name, '12af3rg')


def test_section_str():
    s = _core.Section('foo')
    nt.assert_true(isinstance(str(s), str))


def _check_cloned_neurites(a, b):

    nt.assert_true(a is not b)
    nt.assert_true(a.root_node is not b.root_node)
    nt.assert_equal(a.type, b.type)
    for aa, bb in zip(a.iter_sections(), b.iter_sections()):
        nt.assert_true(np.all(aa.points == bb.points))


def test_neuron_deepcopy():

    d = _io.load_neuron(FILENAMES[0])
    dc = deepcopy(d)

    nt.assert_true(d is not dc)

    nt.assert_true(d.soma is not dc.soma)

    nt.assert_true(np.all(d.soma.points == dc.soma.points))
    nt.assert_true(np.all(d.soma.center == dc.soma.center))
    nt.assert_equal(d.soma.radius, dc.soma.radius)

    for a, b in zip(d.neurites, dc.neurites):
        _check_cloned_neurites(a, b)


def test_neurite_deepcopy():

    d = _io.load_neuron(FILENAMES[0])
    nrt = d.neurites[0]
    nrt2 = deepcopy(nrt)

    nt.assert_true(nrt is not nrt2)

    _check_cloned_neurites(nrt, nrt2)
